package cz.mendelu.dnaAnalyser.analyse.g4hunter;

import cz.mendelu.dnaAnalyser.jwt.JwtAuthenticationUtils;
import cz.mendelu.dnaAnalyser.jwt.JwtTokenService;
import cz.mendelu.dnaAnalyser.sequence.SequenceService;
import cz.mendelu.dnaAnalyser.utils.controller.DeleteMapping;
import cz.mendelu.dnaAnalyser.utils.controller.GetMapping;
import cz.mendelu.dnaAnalyser.utils.controller.*;
import cz.mendelu.dnaAnalyser.utils.controller.PostMapping;
import cz.mendelu.dnaAnalyser.utils.controller.PutMapping;
import cz.mendelu.dnaAnalyser.utils.swagger.ApiPaginationParams;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.Data;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.Authentication;
import org.springframework.web.bind.annotation.*;
import springfox.documentation.annotations.ApiIgnore;

import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

@RestController
@Api(tags = "api/analyse/g4hunter")
@RequestMapping("api/analyse/g4hunter")
public class G4HunterController {

    @Data
    static class G4HunterCreateRequest {

        private UUID sequence;

        private Set<String> tags;

        @NotNull
        private Integer windowSize = 10;

        @NotNull
        private float threshold;
    }

    @Autowired
    private G4HunterService g4HunterService;

    @Autowired
    private SequenceService sequenceService;

    @Autowired
    private JwtTokenService jwtTokenService;

    @GetMapping
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get page with G4hunter analysis.")
    public ResponsePage<G4Hunter> getPageWithG4HunterAnalysis(
            PaginationParam paginationParam,
            SortParam sortParam,
            TagFilterParam tagFilterParam
    ) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<G4Hunter> g4HunterAnalysePage = g4HunterService.findAll(pageable, tagFilterParam.getTags());
        ResponsePage<G4Hunter> responsePage = ResponsePage.responsePage(g4HunterAnalysePage);
        g4HunterAnalysePage
                .forEach(g4Hunter -> responsePage.setDownloadToken(jwtTokenService.generateDownloadToken()));
        return responsePage;
    }

    @GetMapping("/tag")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get all tags defined for sequences")
    public ResponsePage<String> getPageWithTags(
            PaginationParam paginationParam,
            SortParam sortParam
    ) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<String> sequencePage = g4HunterService.findAllTags(pageable);
        return ResponsePage.responsePage(sequencePage);
    }

    @GetMapping("/{id}")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get one G4hunter analyse by ID.")
    public Response<G4Hunter> getOneG4HunterAnalyse(
            @PathVariable UUID id) {
        G4Hunter g4Hunter = g4HunterService.findOne(id);
        Response<G4Hunter> response = Response.response(g4Hunter);
        response.setDownloadToken(jwtTokenService.generateDownloadToken());
        return response;
    }


    @PostMapping()
    @ResponseStatus(HttpStatus.CREATED)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Create G4Hunter analyze.")
    public Response<G4Hunter> createG4Hunter(
            @ApiIgnore Authentication authentication,
            @RequestBody @Valid G4HunterCreateRequest request) {
        G4Hunter g4Hunter = g4HunterService.executeG4HunterAnalyse(
                G4Hunter.builder()
                        .sequence(sequenceService.findOne(request.getSequence()))
                        .owner(JwtAuthenticationUtils.toUser(authentication))
                        .tags(request.getTags())
                        .windowSize(request.getWindowSize())
                        .threshold(request.getThreshold())
                        .build()
        );
        return Response.response(g4Hunter);
    }

    @PutMapping("/{id}/tags")
    @ResponseStatus(HttpStatus.ACCEPTED)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Modify tags")
    public Response<G4Hunter> modifyTags(
            @PathVariable UUID id,
            @RequestBody TagModifyRequest request) {
        G4Hunter analyse = g4HunterService.modifyTags(id, request.getTags());
        return Response.response(analyse);
    }

    @DeleteMapping("/{id}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Delete one G4hunter analyse by ID")
    public void deleteG4HunterAnalyse(
            @PathVariable UUID id) {
        g4HunterService.delete(id);
    }

}
