package cz.mendelu.dnaAnalyser.sequence.stream

import cz.mendelu.dnaAnalyser.sequence.nucleotide.Nucleotide
import spock.lang.Specification
import spock.lang.Unroll

import java.nio.ByteBuffer
import java.util.function.Consumer

/**
 * @author Ing. Jan Kolomazník, Ph.D.
 */
class WindowTest extends Specification {

    def "Create Windows with buffer without buffer.mark()"() {
        setup:
        def buf = ByteBuffer.wrap("ACTGACTG".bytes)

        when:
        BufferedWindow.wrap(buf)

        then:
        thrown(WindowStreamException.class)
    }

    def "getSize() simple test."() {
        setup:
        def window = BufferedWindow.wrap("ACTGACTG")

        expect:
        8 == window.getSize()
    }

    @Unroll
    def "getSize() for #buffer"() {
        setup:
        def window = BufferedWindow.wrap(buffer)

        expect:
        length == window.getSize()

        where:
        length | buffer
        8      | ByteBuffer.wrap("ACTGACTG".bytes).mark()
        4      | ByteBuffer.wrap("ACTGACTG".bytes).position(2).mark().limit(6)
    }

    def "toPlain() simple test."() {
        setup:
        def window = BufferedWindow.wrap("ACTGACTG")

        expect:
        "ACTGACTG" == window.toPlain()
    }

    @Unroll
    def "toPlain() using mark and limit on ByteBuffer #buffer"() {
        setup:
        def window = BufferedWindow.wrap(buffer)

        expect:
        plain == window.toPlain()

        where:
        plain      | buffer
        "ACTGACTG" | ByteBuffer.wrap("ACTGACTG".bytes).mark()
        "TGAC"     | ByteBuffer.wrap("ACTGACTG".bytes).position(2).mark().limit(6)
    }

    def "getPosition() simple test."() {
        setup:
        def window = BufferedWindow.wrap("ACTGACTG")

        expect:
        0 == window.getPosition()
    }

    @Unroll
    def "getPosition() using mark and limit on ByteBuffer #buffer"() {
        setup:
        def window = BufferedWindow.wrap(buffer)

        expect:
        position == window.getPosition()

        where:
        position | buffer
        0        | ByteBuffer.wrap("ACTGACTG".bytes).mark()
        2        | ByteBuffer.wrap("ACTGACTG".bytes).position(2).mark().limit(6)
    }

    def "get(#index) simple test."() {
        setup:
        def window = BufferedWindow.wrap("ACTGACTG")

        expect:
        0 == window.getPosition()
    }

    @Unroll
    def "get(#index) using mark and limit on ByteBuffer #buffer"() {
        setup:
        def window = BufferedWindow.wrap(buffer)

        expect:
        nucleotide == window.get(index)

        where:
        index | nucleotide   | buffer
        0     | Nucleotide.A | ByteBuffer.wrap("ACTGACTG".bytes).mark()
        2     | Nucleotide.T | ByteBuffer.wrap("ACTGACTG".bytes).mark()
        0     | Nucleotide.G | ByteBuffer.wrap("ACTGACTG".bytes).position(3).mark().limit(8)
        2     | Nucleotide.C | ByteBuffer.wrap("ACTGACTG".bytes).position(3).mark().limit(8)
    }

    def "get(index) out of range throw IndexOutOfBoundsException"() {
        setup:
        def window = BufferedWindow.wrap(ByteBuffer.wrap("ACTGACTG".bytes).mark())

        when:
        window.get(10)

        then:
        thrown(IndexOutOfBoundsException)
    }

    def "forEach() simple test."() {
        setup:
        def window = BufferedWindow.wrap("ACTGACTG")
        def consumer = Mock(Consumer)

        when:
        window.forEach(consumer)

        then:
        8 * consumer.accept(_)
    }

    @Unroll
    def "forEach() using mark and limit on ByteBuffer #buffer"() {
        setup:
        def window = BufferedWindow.wrap(buffer)
        def consumer = Mock(Consumer)

        when:
        window.forEach(consumer)

        then:
        count * consumer.accept(_)

        where:
        count | buffer
        8     | ByteBuffer.wrap("ACTGACTG".bytes).mark()
        4     | ByteBuffer.wrap("ACTGACTG".bytes).position(2).mark().limit(6)
    }

    def "doubleStream() simple test."() {
        setup:
        def window = BufferedWindow.wrap("ACTGACTG")
        def consumer = Mock(Consumer)

        when:
        window.doubleStream { i, n -> Byte.valueOf(n.toByte()).doubleValue() }
                .forEach { d -> consumer.accept(d) }
        then:
        8 * consumer.accept(_)
    }

    def "doubleStream() implements Spliterator.OfDouble, test it't method estimateSize() "() {
        setup:
        def window = BufferedWindow.wrap("ACTGACTG")

        when: "Transfer to stream of index (0>1>2>3>4>5>6>7) and count average = 3.375"
        def avg = window.doubleStream { i, n -> Integer.valueOf(i).toDouble() }
                .average()
                .getAsDouble()

        then:
        avg == 3.5
    }
}
